#!/usr/bin/env python3
import os
import json
import subprocess
import shutil
import time
import paramiko
import warnings
from datetime import datetime
from argparse import ArgumentParser
from ftplib import FTP, error_perm, error_temp
from webdav3.client import Client as WebDavClient
from pod2gen import Podcast, Episode, Media, Category, Person


warnings.filterwarnings("ignore", category=UserWarning)

def log_message(message):
    current_time = datetime.now().strftime("[%d.%m.%Y, %H:%M:%S]")
    print(f"{current_time} {message}")

def load_config():
    dir_path = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(dir_path, 'config.json')
    with open(config_path, 'r') as file:
        return json.load(file)

def load_stations():
    dir_path = os.path.dirname(os.path.abspath(__file__))
    stations_path = os.path.join(dir_path, 'stations.json')
    with open(stations_path, 'r') as file:
        return json.load(file)

def normalize_filename(name):
    mapping = str.maketrans('ąćęłńóśźżĄĆĘŁŃÓŚŹŻ', 'acelnoszzACELNOSZZ')
    normalized_name = name.translate(mapping)
    normalized_name = ''.join(c for c in normalized_name if c.isalnum() or c in [' ', '-', '.'])
    return normalized_name

def manage_local_space(directory, config):
    local_max_files = config['local_audio_config']['local_max_files']
    if local_max_files == 0:
        return

    files = [os.path.join(directory, f) for f in os.listdir(directory) if os.path.isfile(os.path.join(directory, f))]
    audio_files = [f for f in files if f.endswith('.mp3') or f.endswith('.opus') or f.endswith('.m4a')]

    while len(audio_files) > local_max_files:
        oldest_file = min(audio_files, key=os.path.getmtime)
        os.remove(oldest_file)
        log_message(f"Deleted oldest local file.")
        audio_files.remove(oldest_file)

def manage_ftp_space(ftp, folder_path, config):
    ftp_max_files = config['ftp_config']['ftp_max_files']
    if ftp_max_files == 0:
        return

    ftp.cwd(folder_path)
    files = ftp.nlst()
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    while len(audio_files) > ftp_max_files:
        oldest_file = min(audio_files, key=lambda x: ftp.sendcmd('MDTM ' + x))
        ftp.delete(oldest_file)
        log_message(f"Deleted oldest file on FTP.")
        audio_files.remove(oldest_file)

def manage_sftp_space(sftp, folder_path, config):
    sftp_max_files = config['sftp_config']['sftp_max_files']
    if sftp_max_files == 0:
        return

    files = sftp.listdir(folder_path)
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    if len(audio_files) <= sftp_max_files:
        return

    files_info = [(file, sftp.stat(os.path.join(folder_path, file)).st_mtime) for file in audio_files]
    files_info.sort(key=lambda x: x[1])

    for file_info in files_info[:-sftp_max_files]:
        sftp.remove(os.path.join(folder_path, file_info[0]))
        log_message(f"Deleted oldest file on SFTP.")

def manage_nextcloud_space(client, folder_path, config):
    nextcloud_max_files = config['nextcloud_config']['nextcloud_max_files']
    if nextcloud_max_files == 0:
        return

    files = client.list(folder_path)
    audio_files = [file for file in files if file.endswith('.mp3') or file.endswith('.opus') or file.endswith('.m4a')]

    if len(audio_files) <= nextcloud_max_files:
        return

    files_info = [(file, client.info(os.path.join(folder_path, file)).get('modified')) for file in audio_files]
    files_info.sort(key=lambda x: x[1])

    for file_info in files_info[:-nextcloud_max_files]:
        client.clean(os.path.join(folder_path, file_info[0]))
        log_message(f"Deleted oldest file on Nextcloud.")

def parse_date_from_filename(filename):
    date_str = filename.split('(')[-1].split(')')[0]
    return datetime.strptime(date_str, '%d.%m.%y %H.%M')

def rss(config, remote_dir, station, show_name, destination):
    if destination == 'ftp':
        web_base_url = config['ftp_config']['web_base_url']
        ftp_host = config['ftp_config']['ftp_hostname']
        ftp_username = config['ftp_config']['ftp_username']
        ftp_password = config['ftp_config']['ftp_password']
    elif destination == 'sftp':
        web_base_url = config['sftp_config']['web_base_url']
        sftp_host = config['sftp_config']['sftp_hostname'].split(':')[0]
        sftp_port = int(config['sftp_config']['sftp_hostname'].split(':')[1])
        sftp_username = config['sftp_config']['sftp_username']
        sftp_password = config['sftp_config']['sftp_password']

    podcast = Podcast()
    podcast.name = f"{show_name} - {station}"
    podcast.authors = [Person(f"{station}")]
    podcast.email = "example@example.com"
    podcast.website = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name)).replace(" ", "%20")
    podcast.description = f"Podcast {show_name} z {station}"
    podcast.language = "pl"
    podcast.category = Category("News", "Politics")
    podcast.explicit = False
    podcast.image = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name), 'podcast.jpg').replace(" ", "%20")

    files_info = []

    if destination == 'ftp':
        with FTP(ftp_host) as ftp:
            ftp.login(ftp_username, ftp_password)
            ftp.cwd(remote_dir)
            ftp.set_pasv(True)
            ftp.sendcmd("TYPE I")
            try:
                items = list(ftp.mlsd())
                for item in items:
                    name, info = item
                    if info.get('type') == 'file' and (name.endswith('.mp3') or name.endswith('.m4a')):
                        size = info.get('size')
                        files_info.append({'name': name, 'size': size})
            except Exception as e:
                log_message(f"Unable to get file info: {e}")

    elif destination == 'sftp':
        try:
            transport = paramiko.Transport((sftp_host, sftp_port))
            transport.connect(username=sftp_username, password=sftp_password)
            sftp = paramiko.SFTPClient.from_transport(transport)
            sftp.chdir(remote_dir)

            for file_attr in sftp.listdir_attr():
                file_name = file_attr.filename
                if file_name.endswith('.mp3') or file_name.endswith('.m4a'):
                    files_info.append({'name': file_name, 'size': file_attr.st_size})

            sftp.close()
            transport.close()

        except Exception as e:
            log_message(f"Error accessing SFTP: {e}")

    web_dir_url = os.path.join(web_base_url, normalize_filename(station), normalize_filename(show_name))

    for file_info in files_info:
        date = parse_date_from_filename(file_info['name'])
        pub_date = date.strftime('%a, %d %b %Y %H:%M:%S GMT')
        episode = Episode()
        episode_title = f"{show_name} - {station} ({date.strftime('%d.%m.%Y')})"
        episode.title = episode_title
        episode.media = Media(url=f"{web_dir_url.replace(' ', '%20')}/{file_info['name'].replace(' ', '%20')}", size=file_info['size'], type="audio/mpeg")
        episode.publication_date = pub_date
        episode.summary = 'Opis odcinka'
        podcast.episodes.append(episode)

    rss_content = podcast.rss_str()

    local_rss_file = 'feed.xml'
    with open(local_rss_file, 'wb') as f:
        f.write(rss_content.encode('utf-8'))

    if destination == 'ftp':
        with FTP(ftp_host) as ftp:
            ftp.login(ftp_username, ftp_password)
            ftp.cwd(remote_dir)
            with open(local_rss_file, 'rb') as f:
                ftp.storbinary(f'STOR {os.path.basename(local_rss_file)}', f)

    elif destination == 'sftp':
        try:
            transport = paramiko.Transport((sftp_host, sftp_port))
            transport.connect(username=sftp_username, password=sftp_password)
            sftp = paramiko.SFTPClient.from_transport(transport)
            sftp.put(local_rss_file, os.path.join(remote_dir, 'feed.xml'))
            sftp.close()
            transport.close()
        except Exception as e:
            log_message(f"Error uploading RSS to SFTP: {e}")

    os.remove(local_rss_file)
    log_message(f"Uploaded RSS file to {destination.upper()}")

def record_show(total_duration, station_url, station, show_name, config, segments=None):
    recordings_path = config['local_audio_config'].get('local_recordings_path', './')
    dir_path = os.path.abspath(recordings_path) if recordings_path != './' else os.path.dirname(os.path.abspath(__file__))

    local_audio_config = config['local_audio_config']
    date_str = datetime.now().strftime("%d.%m.%y %H.%M")

    output_dir = os.path.join(dir_path, "recordings", normalize_filename(station), normalize_filename(show_name))
    os.makedirs(output_dir, exist_ok=True)

    temp_files = []

    if segments:
        # Realizujemy segmenty zdefiniowane w -p (parzyste nagrywanie, nieparzyste przerwy)
        segments_in_seconds = [x * 60 for x in segments]
    else:
        # Jeśli nie podano -p, nagrywamy ciągle przez total_duration
        segments_in_seconds = [total_duration * 60]

    for index in range(0, len(segments_in_seconds), 2):
        temp_filename = os.path.join(output_dir, f"{normalize_filename(show_name)}_{index}.ts")
        temp_files.append(temp_filename)

        # Nagrywanie segmentu
        record_duration = segments_in_seconds[index]
        ffmpeg_command = [
            "ffmpeg",
            "-t", str(record_duration),
            "-reconnect", "1",
            "-reconnect_streamed", "1",
            "-reconnect_at_eof", "1",
            "-reconnect_delay_max", "2",
            "-i", station_url,
            "-c", "copy",  # Kopiowanie bez transkodowania
            "-f", "mpegts",  # Format TS
            temp_filename
        ]
        subprocess.run(ffmpeg_command, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        log_message(f"Nagrywanie trwało {segments_in_seconds[index] // 60} minut: {temp_filename}")

        # Przerwa, jeśli istnieje
        if index + 1 < len(segments_in_seconds):
            log_message(f"Przerwa na {segments_in_seconds[index + 1] // 60} minut.")
            time.sleep(segments_in_seconds[index + 1])

    # Sklejanie plików TS
    combined_ts_file = os.path.join(output_dir, f"{normalize_filename(show_name)} - {normalize_filename(station)} ({date_str}).ts")
    
    with open(os.path.join(output_dir, 'filelist.txt'), 'w') as f:
        for temp_file in temp_files:
            f.write(f"file '{temp_file}'\n")

    ffmpeg_concat_command = [
        "ffmpeg",
        "-f", "concat",
        "-safe", "0",
        "-i", os.path.join(output_dir, 'filelist.txt'),
        "-c", "copy",
        combined_ts_file
    ]
    subprocess.run(ffmpeg_concat_command, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

    log_message(f"Pliki TS zostały połączone: {combined_ts_file}")

    # Usuwanie tymczasowych plików .ts i pliku listy
    for temp_file in temp_files:
        os.remove(temp_file)
    os.remove(os.path.join(output_dir, 'filelist.txt'))

    return combined_ts_file


def convert_audio(source_file, config, destination):
    """
    Konwertuje nagrany plik .ts do formatu docelowego (opus, mp3, etc.) według konfiguracji.
    """
    if destination == 'ftp':
        target_config = config['ftp_config']
    elif destination == 'nextcloud':
        target_config = config['nextcloud_config']
    elif destination == 'sftp':
        target_config = config['sftp_config']
    elif destination == 'local':
        target_config = config['local_audio_config']
    else:
        raise ValueError(f"Unknown destination: {destination}")

    target_format = target_config[f'{destination}_audio_format']
    target_bitrate = target_config[f'{destination}_audio_bitrate']
    channels = target_config[f'{destination}_audio_channels']

    # Nazwa docelowego pliku po konwersji z odpowiednim przedrostkiem, np. _local, _sftp, _nextcloud
    base_filename = os.path.splitext(source_file)[0]
    target_filename = base_filename + f"_{destination}.{target_format}"
    log_message(f"Konwertowanie pliku do formatu {target_format.upper()} dla {destination.upper()}.")

    # Komenda do konwersji z .ts do odpowiedniego formatu
    ffmpeg_command = [
        "ffmpeg",
        "-i", source_file,
        "-c:a", "libopus" if target_format == "opus" else "libmp3lame" if target_format == "mp3" else "copy",
        "-b:a", f"{target_bitrate}k",  # Ustawienie bitrate
        "-ac", str(channels),  # Ustawienie liczby kanałów
        "-y", target_filename
    ]
    subprocess.run(ffmpeg_command, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

    log_message(f"Konwersja zakończona: {target_filename}")
    return target_filename

def upload_ftp(filename, config, station, show_name):
    attempts = 0
    success = False
    while not success and attempts < 5:
        try:
            with FTP(config['ftp_config']['ftp_hostname']) as ftp:
                ftp.login(config['ftp_config']['ftp_username'], config['ftp_config']['ftp_password'])
                remote_dir = f"{config['ftp_config']['ftp_base_path']}/{normalize_filename(station)}/{normalize_filename(show_name)}"

                parts = remote_dir.split('/')
                current_dir = ''
                for part in parts:
                    if part:
                        current_dir += f'/{part}'
                        try:
                            ftp.cwd(current_dir)
                        except error_perm as e:
                            if 'No such file or directory' in str(e):
                                ftp.mkd(current_dir)
                                ftp.cwd(current_dir)
                            else:
                                raise

                with open(filename, 'rb') as file:
                    ftp.storbinary('STOR ' + os.path.basename(filename), file)
                success = True
                manage_ftp_space(ftp, remote_dir, config)
                log_message(f"Uploaded file to FTP.")
                rss(config, remote_dir, station, show_name, 'ftp')
        except (error_temp, BrokenPipeError) as e:
            attempts += 1
            log_message(f"Error: {e}. Retrying... ({attempts})")
        except Exception as e:
            log_message(f"Unexpected error: {e}")
            break
    if success:
        os.remove(filename)  # Usunięcie pliku po uploadzie
        log_message(f"Deleted local file after FTP upload: {filename}")
    if not success:
        log_message(f"Failed to upload {filename} to FTP after several attempts.")

def upload_sftp(filename, config, station, show_name):
    attempts = 0
    success = False
    while not success and attempts < 5:
        try:
            sftp_config = config['sftp_config']
            transport = paramiko.Transport((sftp_config['sftp_hostname'].split(':')[0], int(sftp_config['sftp_hostname'].split(':')[1])))
            transport.connect(username=sftp_config['sftp_username'], password=sftp_config['sftp_password'])
            sftp = paramiko.SFTPClient.from_transport(transport)

            remote_dir = f"{sftp_config['sftp_base_path']}/{normalize_filename(station)}/{normalize_filename(show_name)}"
            try:
                sftp.chdir(remote_dir)
            except IOError:
                parts = remote_dir.split('/')
                current_dir = ''
                for part in parts:
                    if part:
                        current_dir += f'/{part}'
                        try:
                            sftp.chdir(current_dir)
                        except IOError:
                            sftp.mkdir(current_dir)
                            sftp.chdir(current_dir)

            sftp.put(filename, os.path.basename(filename))
            success = True
            log_message(f"Uploaded file to SFTP.")
            manage_sftp_space(sftp, remote_dir, config)
            rss(config, remote_dir, station, show_name, 'sftp')
        except Exception as e:
            attempts += 1
            log_message(f"Error: {e}. Retrying... ({attempts})")
        finally:
            sftp.close()
            transport.close()
    if success:
        os.remove(filename)  # Usunięcie pliku po uploadzie
        log_message(f"Deleted local file after SFTP upload: {filename}")
    if not success:
        log_message(f"Failed to upload {filename} to SFTP after several attempts.")

def upload_nextcloud(filename, config, station, show_name):
    options = {
        'webdav_hostname': config['nextcloud_config']['nextcloud_hostname'],
        'webdav_login': config['nextcloud_config']['nextcloud_login'],
        'webdav_password': config['nextcloud_config']['nextcloud_password']
    }
    client = WebDavClient(options)
    remote_dir = f"/public/{normalize_filename(station)}/{normalize_filename(show_name)}/"
    remote_path = remote_dir + os.path.basename(filename)

    ensure_remote_path(client, remote_path)
    client.upload_sync(remote_path=remote_path, local_path=filename)
    log_message(f"Uploaded file to NextCloud.")
    manage_nextcloud_space(client, remote_dir, config)
    os.remove(filename)
    log_message(f"Deleted local file after NextCloud upload: {filename}")

def ensure_remote_path(client, remote_path):
    parts = remote_path.strip('/').split('/')
    path = ''
    for part in parts[:-1]:
        path += f"/{part}"
        if not client.check(path):
            client.mkdir(path)

def main():
    parser = ArgumentParser()
    parser.add_argument("station", type=str, help="Station name")
    parser.add_argument("show_name", type=str, help="Name of the show")
    parser.add_argument("total_duration", type=int, help="Total duration of the recording in minutes")
    parser.add_argument("-p", "--pauses", type=str, help="Sequence of recording and pause times, e.g., '10,2,8,1'")
    parser.add_argument("-s", "--save", nargs='+', choices=['ftp', 'nextcloud', 'local', 'sftp'], help="Save destinations", default=[])
    args = parser.parse_args()

    config = load_config()
    stations = load_stations()
    station_url = stations[args.station]

    # Parsowanie wartości -p
    if args.pauses:
        segments = list(map(int, args.pauses.split(',')))
    else:
        segments = [args.total_duration]  # Jeśli nie podano przerw, nagrywamy cały czas

    # Wywołanie funkcji record_show z odpowiednimi argumentami
    combined_ts_file = record_show(args.total_duration, station_url, args.station, args.show_name, config, segments)

    # Konwertujemy plik .ts do każdego formatu z pojedynczym przedrostkiem
    if 'local' in args.save:
        local_file = convert_audio(combined_ts_file, config, 'local')
        manage_local_space(os.path.dirname(local_file), config)

    if 'ftp' in args.save:
        ftp_file = convert_audio(combined_ts_file, config, 'ftp')
        upload_ftp(ftp_file, config, args.station, args.show_name)

    if 'nextcloud' in args.save:
        nextcloud_file = convert_audio(combined_ts_file, config, 'nextcloud')
        upload_nextcloud(nextcloud_file, config, args.station, args.show_name)

    if 'sftp' in args.save:
        sftp_file = convert_audio(combined_ts_file, config, 'sftp')
        upload_sftp(sftp_file, config, args.station, args.show_name)

    # Usuwamy plik .ts po zakończeniu wszystkich operacji
    os.remove(combined_ts_file)

    log_message("Recording and uploads complete.")

if __name__ == "__main__":
    main()